@@ -207,9 +207,6 @@
 #define USE_ICONV (tds->conn->use_iconv)
 
 static const TDSCOLUMNFUNCS *tds_get_column_funcs(TDSCONNECTION *conn, int type);
-#ifdef WORDS_BIGENDIAN
-static void tds_swap_datatype(int coltype, void *b);
-#endif
 static void tds_swap_numeric(TDS_NUMERIC *num);
 
 #undef MIN
@@ -401,7 +398,11 @@ tds_generic_get_info(TDSSOCKET *tds, TDSCOLUMN *col)
 		col->column_size = tds_get_smallint(tds);
 		/* under TDS7.2 this means ?var???(MAX) */
 		if (col->column_size < 0 && IS_TDS72_PLUS(tds->conn)) {
-			col->column_size = 0x3ffffffflu;
+			if (is_char_type(col->column_type))
+				col->column_size = 0x3ffffffflu;
+			else
+				col->column_size = 0x7ffffffflu;
+
 			col->column_varint_size = 8;
 		}
 		if (col->column_size < 0)
@@ -442,7 +443,7 @@ tds_generic_get_info(TDSSOCKET *tds, TDSCOLUMN *col)
 	} else if (IS_TDS72_PLUS(tds->conn) && col->on_server.column_type == SYBMSXML) {
 		unsigned char has_schema = tds_get_byte(tds);
 		if (has_schema) {
-			/* discard schema informations */
+			/* discard schema information */
 			tds_get_string(tds, tds_get_byte(tds), NULL, 0);        /* dbname */
 			tds_get_string(tds, tds_get_byte(tds), NULL, 0);        /* schema owner */
 			tds_get_string(tds, tds_get_usmallint(tds), NULL, 0);    /* schema collection */
@@ -474,9 +475,7 @@ tds_get_char_dynamic(TDSSOCKET *tds, TDSCOLUMN *curcol, void **pp, size_t alloca
 	 * Blobs don't use a column's fixed buffer because the official maximum size is 2 GB.
 	 * Instead, they're reallocated as necessary, based on the data's size.
 	 */
-	res = tds_dynamic_stream_init(&w, pp, allocated);
-	if (TDS_FAILED(res))
-		return res;
+	TDS_PROPAGATE(tds_dynamic_stream_init(&w, pp, allocated));
 
 	if (USE_ICONV && curcol->char_conv)
 		res = tds_convert_stream(tds, curcol->char_conv, to_client, r_stream, &w.stream);
@@ -579,7 +578,7 @@ tds_variant_get(TDSSOCKET * tds, TDSCOLUMN * curcol)
 
 	type = (TDS_SERVER_TYPE) tds_get_byte(tds);
 	info_len = tds_get_byte(tds);
-	if (!is_tds_type_valid(type))
+	if (!is_variant_inner_type(type))
 		goto error_type;
 	v = (TDSVARIANT*) curcol->column_data;
 	v->type = type;
@@ -605,6 +604,8 @@ tds_variant_get(TDSSOCKET * tds, TDSCOLUMN * curcol)
 			TDS_ZERO_FREE(v->data);
 		v->data_len = sizeof(TDS_NUMERIC);
 		num = tds_new0(TDS_NUMERIC, 1);
+		if (!num)
+			goto error_memory;
 		v->data = (TDS_CHAR *) num;
 		num->precision = tds_get_byte(tds);
 		num->scale     = tds_get_byte(tds);
@@ -675,20 +676,16 @@ tds_variant_get(TDSSOCKET * tds, TDSCOLUMN * curcol)
 	if (v->data)
 		TDS_ZERO_FREE(v->data);
 	if (colsize) {
-		TDSRET res;
 		TDSDATAINSTREAM r;
 
 		if (USE_ICONV && curcol->char_conv)
 			v->type = tds_get_cardinal_type(type, 0);
 
 		tds_datain_stream_init(&r, tds, colsize);
-		res = tds_get_char_dynamic(tds, curcol, (void **) &v->data, colsize, &r.stream);
-		if (TDS_FAILED(res))
-			return res;
+		TDS_PROPAGATE(tds_get_char_dynamic(tds, curcol, (void **) &v->data, colsize, &r.stream));
 		colsize = curcol->column_cur_size;
 #ifdef WORDS_BIGENDIAN
-		if (tds->conn->emul_little_endian)
-			tds_swap_datatype(tds_get_conversion_type(type, colsize), v->data);
+		tds_swap_datatype(tds_get_conversion_type(type, colsize), v->data);
 #endif
 	}
 	v->data_len = colsize;
@@ -696,6 +693,7 @@ tds_variant_get(TDSSOCKET * tds, TDSCOLUMN * curcol)
 	return TDS_SUCCESS;
 
 error_type:
+error_memory:
 	tds_get_n(tds, NULL, colsize);
 	return TDS_FAIL;
 }
@@ -726,10 +724,10 @@ tds_generic_get(TDSSOCKET * tds, TDSCOLUMN * curcol)
 		if (len == 16) {	/*  Jeff's hack */
 			tds_get_n(tds, blob->textptr, 16);
 			tds_get_n(tds, blob->timestamp, 8);
-			blob->valid_ptr = 1;
+			blob->valid_ptr = true;
 			if (IS_TDS72_PLUS(tds->conn) &&
 			    memcmp(blob->textptr, "dummy textptr\0\0",16) == 0)
-				blob->valid_ptr = 0;
+				blob->valid_ptr = false;
 			colsize = tds_get_int(tds);
 		} else {
 			colsize = -1;
@@ -813,8 +811,7 @@ tds_generic_get(TDSSOCKET * tds, TDSCOLUMN * curcol)
 	/* non-numeric and non-blob */
 
 	if (USE_ICONV && curcol->char_conv) {
-		if (TDS_FAILED(tds_get_char_data(tds, (char *) dest, colsize, curcol)))
-			return TDS_FAIL;
+		TDS_PROPAGATE(tds_get_char_data(tds, (char *) dest, colsize, curcol));
 	} else {
 		/*
 		 * special case, some servers seem to return more data in some conditions
@@ -856,10 +853,8 @@ tds_generic_get(TDSSOCKET * tds, TDSCOLUMN * curcol)
 	}
 
 #ifdef WORDS_BIGENDIAN
-	if (tds->conn->emul_little_endian) {
-		tdsdump_log(TDS_DBG_INFO1, "swapping coltype %d\n", tds_get_conversion_type(curcol->column_type, colsize));
-		tds_swap_datatype(tds_get_conversion_type(curcol->column_type, colsize), dest);
-	}
+	tdsdump_log(TDS_DBG_INFO1, "swapping coltype %d\n", tds_get_conversion_type(curcol->column_type, colsize));
+	tds_swap_datatype(tds_get_conversion_type(curcol->column_type, colsize), dest);
 #endif
 	return TDS_SUCCESS;
 }
@@ -908,34 +903,6 @@ tds_generic_put_info(TDSSOCKET * tds, TDSCOLUMN * col)
 	return TDS_SUCCESS;
 }
 
-unsigned
-tds_generic_put_info_len(TDSSOCKET * tds, TDSCOLUMN * col)
-{
-	unsigned len = col->column_varint_size;
-
-	CHECK_TDS_EXTRA(tds);
-	CHECK_COLUMN_EXTRA(col);
-
-	switch (col->column_varint_size) {
-	case 5:
-		len = 4;
-		break;
-	case 8:
-		len = 2;
-		break;
-	}
-
-	if (IS_TDS50(tds->conn)
-	    && (col->on_server.column_type == SYBIMAGE || col->on_server.column_type == SYBTEXT))
-		len += 2;
-
-	/* TDS7.1 output collate information */
-	if (IS_TDS71_PLUS(tds->conn) && is_collate_type(col->on_server.column_type))
-		len += 5;
-
-	return len;
-}
-
 /**
  * Write data to wire
  * \param tds state information for the socket and the TDS protocol
@@ -1035,7 +1002,7 @@ tds_generic_put(TDSSOCKET * tds, TDSCOLUMN * curcol, int bcp7)
 			/* this difference for BCP operation is due to
 			 * a bug in different server version that does
 			 * not accept a length here */
-			tds_put_int8(tds, bcp7 ? -2 : colsize);
+			tds_put_int8(tds, bcp7 ? (TDS_INT8) -2 : (TDS_INT8) colsize);
 			tds_put_int(tds, colsize);
 			break;
 		case 4:	/* It's a BLOB... */
@@ -1077,7 +1044,7 @@ tds_generic_put(TDSSOCKET * tds, TDSCOLUMN * curcol, int bcp7)
 #ifdef WORDS_BIGENDIAN
 			unsigned char buf[64];
 
-			if (tds->conn->emul_little_endian && !converted && colsize < 64) {
+			if (!converted && colsize < 64) {
 				tdsdump_log(TDS_DBG_INFO1, "swapping coltype %d\n",
 					    tds_get_conversion_type(curcol->column_type, colsize));
 				memcpy(buf, s, colsize);
@@ -1140,7 +1107,7 @@ tds_generic_put(TDSSOCKET * tds, TDSCOLUMN * curcol, int bcp7)
 #ifdef WORDS_BIGENDIAN
 			unsigned char buf[64];
 
-			if (tds->conn->emul_little_endian && !converted && colsize < 64) {
+			if (!converted && colsize < 64) {
 				tdsdump_log(TDS_DBG_INFO1, "swapping coltype %d\n",
 					    tds_get_conversion_type(curcol->column_type, colsize));
 				memcpy(buf, s, colsize);
@@ -1172,7 +1139,7 @@ tds_numeric_get_info(TDSSOCKET *tds, TDSCOLUMN *col)
 }
 
 TDS_INT
-tds_numeric_row_len(TDSCOLUMN *col)
+tds_numeric_row_len(TDSCOLUMN *col TDS_UNUSED)
 {
 	return sizeof(TDS_NUMERIC);
 }
@@ -1240,17 +1207,8 @@ tds_numeric_put_info(TDSSOCKET * tds, TDSCOLUMN * col)
 	return TDS_SUCCESS;
 }
 
-unsigned
-tds_numeric_put_info_len(TDSSOCKET * tds, TDSCOLUMN * col)
-{
-	CHECK_TDS_EXTRA(tds);
-	CHECK_COLUMN_EXTRA(col);
-
-	return 3;
-}
-
 TDSRET
-tds_numeric_put(TDSSOCKET *tds, TDSCOLUMN *col, int bcp7)
+tds_numeric_put(TDSSOCKET *tds, TDSCOLUMN *col, int bcp7 TDS_UNUSED)
 {
 	TDS_NUMERIC *num = (TDS_NUMERIC *) col->column_data, buf;
 	unsigned char colsize;
@@ -1270,14 +1228,14 @@ tds_numeric_put(TDSSOCKET *tds, TDSCOLUMN *col, int bcp7)
 }
 
 TDSRET
-tds_variant_put_info(TDSSOCKET * tds, TDSCOLUMN * col)
+tds_variant_put_info(TDSSOCKET * tds TDS_UNUSED, TDSCOLUMN * col TDS_UNUSED)
 {
 	/* TODO */
 	return TDS_FAIL;
 }
 
 TDSRET
-tds_variant_put(TDSSOCKET *tds, TDSCOLUMN *col, int bcp7)
+tds_variant_put(TDSSOCKET *tds TDS_UNUSED, TDSCOLUMN *col TDS_UNUSED, int bcp7 TDS_UNUSED)
 {
 	/* TODO */
 	return TDS_FAIL;
@@ -1297,7 +1255,7 @@ tds_msdatetime_get_info(TDSSOCKET * tds, TDSCOLUMN * col)
 }
 
 TDS_INT
-tds_msdatetime_row_len(TDSCOLUMN *col)
+tds_msdatetime_row_len(TDSCOLUMN *col TDS_UNUSED)
 {
 	return sizeof(TDS_DATETIMEALL);
 }
@@ -1377,7 +1335,7 @@ tds_msdatetime_put_info(TDSSOCKET * tds, TDSCOLUMN * col)
 }
 
 TDSRET
-tds_msdatetime_put(TDSSOCKET *tds, TDSCOLUMN *col, int bcp7)
+tds_msdatetime_put(TDSSOCKET *tds, TDSCOLUMN *col, int bcp7 TDS_UNUSED)
 {
 	const TDS_DATETIMEALL *dta = (const TDS_DATETIMEALL *) col->column_data;
 	unsigned char buf[12], *p;
@@ -1444,16 +1402,10 @@ tds_clrudt_row_len(TDSCOLUMN *col)
 	return sizeof(TDSBLOB);
 }
 
-unsigned
-tds_clrudt_put_info_len(TDSSOCKET * tds, TDSCOLUMN * col)
-{
-	return 3;
-}
-
 TDSRET
-tds_clrudt_put_info(TDSSOCKET * tds, TDSCOLUMN * col)
+tds_clrudt_put_info(TDSSOCKET * tds, TDSCOLUMN * col TDS_UNUSED)
 {
-	/* FIXME support properly*/
+	/* FIXME support properly */
 	tds_put_byte(tds, 0);	/* db_name */
 	tds_put_byte(tds, 0);	/* schema_name */
 	tds_put_byte(tds, 0);	/* type_name */
@@ -1472,7 +1424,7 @@ tds_sybbigtime_get_info(TDSSOCKET * tds, TDSCOLUMN * col)
 }
 
 TDS_INT
-tds_sybbigtime_row_len(TDSCOLUMN *col)
+tds_sybbigtime_row_len(TDSCOLUMN *col TDS_UNUSED)
 {
 	return sizeof(TDS_UINT8);
 }
@@ -1495,21 +1447,15 @@ tds_sybbigtime_get(TDSSOCKET * tds, TDSCOLUMN * col)
 }
 
 TDSRET
-tds_sybbigtime_put_info(TDSSOCKET * tds, TDSCOLUMN * col)
+tds_sybbigtime_put_info(TDSSOCKET * tds, TDSCOLUMN * col TDS_UNUSED)
 {
 	tds_put_byte(tds, 8);
 	tds_put_byte(tds, 6);
 	return TDS_SUCCESS;
 }
 
-unsigned
-tds_sybbigtime_put_info_len(TDSSOCKET * tds, TDSCOLUMN * col)
-{
-	return 2;
-}
-
 TDSRET
-tds_sybbigtime_put(TDSSOCKET *tds, TDSCOLUMN *col, int bcp7)
+tds_sybbigtime_put(TDSSOCKET *tds, TDSCOLUMN *col, int bcp7 TDS_UNUSED)
 {
 	const TDS_UINT8 *dt = (const TDS_UINT8 *) col->column_data;
 
@@ -1525,44 +1471,134 @@ tds_sybbigtime_put(TDSSOCKET *tds, TDSCOLUMN *col, int bcp7)
 }
 
 TDSRET
-tds_invalid_get_info(TDSSOCKET * tds, TDSCOLUMN * col)
+tds_mstabletype_get_info(TDSSOCKET *tds TDS_UNUSED, TDSCOLUMN *col TDS_UNUSED)
+{
+	/* Table type is strictly only an input variable */
+	return TDS_FAIL;
+}
+
+TDS_INT
+tds_mstabletype_row_len(TDSCOLUMN *col TDS_UNUSED)
+{
+	return sizeof(TDS_TVP);
+}
+
+TDSRET
+tds_mstabletype_get(TDSSOCKET *tds TDS_UNUSED, TDSCOLUMN *col TDS_UNUSED)
+{
+	/* Table type is strictly only an input variable */
+	return TDS_FAIL;
+}
+
+TDSRET
+tds_mstabletype_put_info(TDSSOCKET *tds, TDSCOLUMN *col)
+{
+	TDS_TVP *table = (TDS_TVP *) col->column_data;
+	TDSFREEZE current_freeze[1];
+	size_t written;
+
+	/* TVP_TYPENAME */
+	tds_put_byte(tds, 0); /* Empty DB name */
+
+	tds_freeze(tds, current_freeze, 1);
+	tds_put_string(tds, table->schema, -1);
+	written = tds_freeze_written(current_freeze) / 2;
+	tds_freeze_close_len(current_freeze, written);
+
+	tds_freeze(tds, current_freeze, 1);
+	tds_put_string(tds, table->name, -1);
+	written = tds_freeze_written(current_freeze) / 2;
+	tds_freeze_close_len(current_freeze, written);
+
+	return TDS_SUCCESS;
+}
+
+TDSRET
+tds_mstabletype_put(TDSSOCKET *tds, TDSCOLUMN *col, int bcp7 TDS_UNUSED)
+{
+	TDS_TVP *table = (TDS_TVP *) col->column_data;
+	TDSPARAMINFO *params;
+	TDSCOLUMN *tds_col;
+	TDS_TVP_ROW *row;
+	int i;
+	TDS_USMALLINT num_cols = table->metadata ? table->metadata->num_cols : 0;
+
+	/* COL_METADATA */
+	if (num_cols == 0)
+		tds_put_smallint(tds, 0xffff); /* TVP_NULL_TOKEN */
+	else {
+		tds_put_smallint(tds, num_cols);
+
+		params = table->metadata;
+		for (i = 0; i < num_cols; i++) {
+			tds_col = params->columns[i];
+
+			/* UserType*/
+			tds_put_int(tds, tds_col->column_usertype);
+			/* Flags */
+			tds_put_smallint(tds, tds_col->column_flags);
+			/* TYPE_INFO */
+			tds_put_byte(tds, tds_col->on_server.column_type);
+			TDS_PROPAGATE(tds_col->funcs->put_info(tds, tds_col));
+
+			/* ColName - Empty string */
+			tds_put_byte(tds, 0x00);
+		}
+	}
+
+	/* TVP_END_TOKEN */
+	tds_put_byte(tds, 0x00);
+
+	for (row = table->row; row != NULL; row = row->next) {
+		/* TVP_ROW_TOKEN */
+		tds_put_byte(tds, 0x01);
+
+		params = row->params;
+		for (i = 0; i < num_cols; i++) {
+			tds_col = params->columns[i];
+			TDS_PROPAGATE(tds_col->funcs->put_data(tds, tds_col, 0));
+		}
+	}
+
+	/* TVP_END_TOKEN */
+	tds_put_byte(tds, 0x00);
+
+	return TDS_SUCCESS;
+}
+
+TDSRET
+tds_invalid_get_info(TDSSOCKET * tds TDS_UNUSED, TDSCOLUMN * col TDS_UNUSED)
 {
 	return TDS_FAIL;
 }
 
 TDS_INT
-tds_invalid_row_len(TDSCOLUMN *col)
+tds_invalid_row_len(TDSCOLUMN *col TDS_UNUSED)
 {
 	return 0;
 }
 
 TDSRET
-tds_invalid_get(TDSSOCKET * tds, TDSCOLUMN * col)
+tds_invalid_get(TDSSOCKET * tds TDS_UNUSED, TDSCOLUMN * col TDS_UNUSED)
 {
 	return TDS_FAIL;
 }
 
 TDSRET
-tds_invalid_put_info(TDSSOCKET * tds, TDSCOLUMN * col)
+tds_invalid_put_info(TDSSOCKET * tds TDS_UNUSED, TDSCOLUMN * col TDS_UNUSED)
 {
 	return TDS_FAIL;
 }
 
-unsigned
-tds_invalid_put_info_len(TDSSOCKET * tds, TDSCOLUMN * col)
-{
-	return 0;
-}
-
 TDSRET
-tds_invalid_put(TDSSOCKET *tds, TDSCOLUMN *col, int bcp7)
+tds_invalid_put(TDSSOCKET *tds TDS_UNUSED, TDSCOLUMN *col TDS_UNUSED, int bcp7 TDS_UNUSED)
 {
 	return TDS_FAIL;
 }
 
 #if ENABLE_EXTRA_CHECKS
 int
-tds_generic_check(const TDSCOLUMN *col)
+tds_generic_check(const TDSCOLUMN *col TDS_UNUSED)
 {
 	return 0;
 }
@@ -1585,7 +1621,13 @@ tds_sybbigtime_check(const TDSCOLUMN *col)
 }
 
 int
-tds_clrudt_check(const TDSCOLUMN *col)
+tds_mstabletype_check(const TDSCOLUMN *col TDS_UNUSED)
+{
+	return 0;
+}
+
+int
+tds_clrudt_check(const TDSCOLUMN *col TDS_UNUSED)
 {
 	return 0;
 }
@@ -1612,7 +1654,7 @@ tds_msdatetime_check(const TDSCOLUMN *col)
 }
 
 int
-tds_variant_check(const TDSCOLUMN *col)
+tds_variant_check(const TDSCOLUMN *col TDS_UNUSED)
 {
 	return 0;
 }
@@ -1634,7 +1676,7 @@ tds_numeric_check(const TDSCOLUMN *col)
 }
 
 int
-tds_invalid_check(const TDSCOLUMN *col)
+tds_invalid_check(const TDSCOLUMN *col TDS_UNUSED)
 {
 	return 1;
 }
@@ -1652,6 +1694,7 @@ TDS_DECLARE_FUNCS(msdatetime);
 TDS_DECLARE_FUNCS(clrudt);
 TDS_DECLARE_FUNCS(sybbigtime);
 TDS_DECLARE_FUNCS(invalid);
+TDS_DECLARE_FUNCS(mstabletype);
 #include <freetds/popvis.h>
 
 static const TDSCOLUMNFUNCS *
@@ -1675,13 +1718,15 @@ tds_get_column_funcs(TDSCONNECTION *conn, int type)
 	case SYB5BIGTIME:
 	case SYB5BIGDATETIME:
 		return &tds_sybbigtime_funcs;
+	case SYBMSTABLE:
+		return &tds_mstabletype_funcs;
 	}
 	return &tds_generic_funcs;
 }
 #include "tds_types.h"
 
 #ifdef WORDS_BIGENDIAN
-static void
+void
 tds_swap_datatype(int coltype, void *b)
 {
 	unsigned char *buf = (unsigned char *) b;
@@ -1690,12 +1735,14 @@ tds_swap_datatype(int coltype, void *b)
 	case SYBDATETIME4:
 		tds_swap_bytes(&buf[2], 2);
 	case SYBINT2:
+	case SYBUINT2:
 		tds_swap_bytes(buf, 2);
 		break;
 	case SYBMONEY:
 	case SYBDATETIME:
 		tds_swap_bytes(&buf[4], 4);
 	case SYBINT4:
+	case SYBUINT4:
 	case SYBMONEY4:
 	case SYBREAL:
 	case SYBDATE:
@@ -1703,6 +1750,7 @@ tds_swap_datatype(int coltype, void *b)
 		tds_swap_bytes(buf, 4);
 		break;
 	case SYBINT8:
+	case SYBUINT8:
 	case SYBFLT8:
 	case SYB5BIGTIME:
 	case SYB5BIGDATETIME:
